# 
# Dynamic features of Python for defining simple DSLs (Proof of Concept Example)
#
# Copyright (C) 2010, Konstantin Tretyakov.
# Licensed under the MIT license.
#
# The following resources were heavily used in the making of this code
#
# http://thermalnoise.wordpress.com/2007/12/30/exploring-python-bytecode/
# http://www.voidspace.org.uk/python/articles/code_blocks.shtml
# http://www.oliverh.com/projects/python-codeutils
# http://code.activestate.com/recipes/498242/
#

def bytecode(code_obj):
	import opcode
	code = code_obj.co_code
	names = code_obj.co_names
	n = len(code)
	i = 0
	while i < n:
		op = ord(code[i])
		i += 1
		oparg = None
		if op >= opcode.HAVE_ARGUMENT:
			oparg = ord(code[i]) + ord(code[i+1])*256
			i += 2
		yield (op, oparg)

def as_anonymous_block(fun):
	'''
	Given a function with no arguments, recodes all local/global variable access in it to use the provided environment,
	essentially turning it into an "anonymous block" and making it usable in the "exec f in env" construction.
	Returns a code object.
	'''
	from types import CodeType
	import opcode
	LOAD_FAST, LOAD_GLOBAL, LOAD_DEREF, LOAD_NAME = map(lambda x: opcode.opmap[x], ['LOAD_FAST', 'LOAD_GLOBAL', 'LOAD_DEREF', 'LOAD_NAME'])
	STORE_FAST, STORE_GLOBAL, STORE_DEREF, STORE_NAME = map(lambda x: opcode.opmap[x], ['STORE_FAST','STORE_GLOBAL','STORE_DEREF','STORE_NAME'])
	
	co = fun.func_code
	len_co_names = len(co.co_names)
	new_co_names = co.co_names + co.co_varnames
	new_co_flags = co.co_flags & ~0x02		# bitmap: 1=optimized | 2=newlocals | 4=*arg | 8=**arg
	new_code = []
	for op, arg in bytecode(fun.func_code):
		if op == STORE_FAST:
			new_arg = arg + len_co_names
			new_code.append(chr(STORE_NAME) + chr(new_arg % 256) + chr(new_arg // 256))
		elif op == LOAD_FAST:
			new_arg = arg + len_co_names
			new_code.append(chr(LOAD_NAME) + chr(new_arg % 256) + chr(new_arg // 256))
		elif op in (LOAD_GLOBAL, LOAD_DEREF):
			new_code.append(chr(LOAD_NAME) + chr(arg % 256) + chr(arg // 256))
		elif op in (STORE_GLOBAL, STORE_DEREF):
			new_code.append(chr(STORE_NAME) + chr(arg % 256) + chr(arg // 256))
		elif arg is None:
			new_code.append(chr(op))
		else:
			new_code.append(chr(op) + chr(arg % 256) + chr(arg // 256))
	new_code = ''.join(new_code)
	return CodeType(co.co_argcount, co.co_nlocals, co.co_stacksize,
                       new_co_flags, new_code, co.co_consts, new_co_names,
                       co.co_varnames, co.co_filename, co.co_name,
                       co.co_firstlineno, co.co_lnotab,
                       co.co_freevars, co.co_cellvars)


class Proxy:
	def __init__(self, name, env):
		self.__dict__['__name__'] = name
		self.__dict__['__env__'] = env
	def __getitem__(self, name):
		return Proxy(str(self) + '[' + str(name) + ']', self.__dict__['__env__'])
	def __setitem__(self, name, value):
		newobj = Proxy(str(self) + '[' + str(name) + ']', self.__dict__['__env__'])
		self.__dict__['__env__'].definitions.append((newobj, value))
	def __getattr__(self, name):
		return Proxy(str(self) + '.' + name, self.__dict__['__env__'])
	def __setattr__(self, name, value):
		newobj = Proxy(str(self) + '.' + name, self.__dict__['__env__'])
		self.__dict__['__env__'].definitions.append((newobj, value))
	def __str__(self):
		return self.__dict__['__name__']
	def __mul__(self,val):
		return Proxy(str(self) + '*' + str(val), self.__dict__['__env__'])
	def __add__(self,val):
		return Proxy(str(self) + '+' + str(val), self.__dict__['__env__'])
	def __coerce__(self, other):
		return None
	def __call__(self, *args, **kwargs):
		a = ','.join(map(str, args))
		kwa = ','.join(map(lambda (x,y): str(x) + '=' + str(y), kwargs.iteritems()))
		args = ','.join([a,kwa])
		return Proxy(str(self) + '(' + args + ')', self.__dict__['__env__'])

class Env(dict):
	def __init__(self):
		self.definitions = []
	def __getitem__(self, name):
		return Proxy(name, self)
	def __setitem__(self, name, value):
		self.definitions.append((name, value))
	def display(self):
		for (lhs, rhs) in self.definitions:
			print str(lhs) + ' := ' + str(rhs)


def my_dsl(f):
	blk = as_anonymous_block(f)
	result = Env()
	exec blk in result
	return result


